"""
from rest_framework import routers
from .views import MyModel1ViewSet


router = routers.DefaultRouter()
router.register('', MyModel1ViewSet)
urlpatterns = router.urls

"""
from django.conf import settings
from packaging import version

NETBOX_CURRENT_VERSION = version.parse(settings.VERSION)
if NETBOX_CURRENT_VERSION >= version.parse("3.3") :
    from netbox.api.routers import NetBoxRouter
elif NETBOX_CURRENT_VERSION >= version.parse("3.2") :
    from netbox.api import NetBoxRouter
else:
    from netbox.api import OrderedDefaultRouter as NetBoxRouter

from . import views

router = NetBoxRouter()

router.register('ic', views.ICViewSet)
router.register('service', views.ServiceViewSet)
router.register('application', views.ApplicationViewSet)

urlpatterns = router.urls
app_name = 'nb_service-api'


from extras.plugins import PluginMenuItem


menu_items = (
    PluginMenuItem(
        link="plugins:nb_service:service_list",
        link_text="Services",
    ),
    PluginMenuItem(
        link="plugins:nb_service:application_list",
        link_text="Applications",
    ),
)



import django_filters
from django.db.models import Q
from tenancy.models import Tenant
from dcim.models import Device
from virtualization.models import VirtualMachine

from . import models

class ICFilter(django_filters.FilterSet):
    
    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    service = django_filters.ModelMultipleChoiceFilter(
        field_name="service__name",
        queryset=models.Service.objects.all(),
        to_field_name="name",
    )
    service_id = django_filters.ModelMultipleChoiceFilter(
        field_name="service__id",
        queryset=models.Service.objects.all(),
        to_field_name="id",
    )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        ids = []
        for ic in queryset:
            if value in ic.name:
                ids.append(ic.id)
        return queryset.filter(id__in = ids)

    class Meta:
        model = models.IC

        fields = [
            "service",
        ]

class ServiceFilter(django_filters.FilterSet):
    
    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    clients = django_filters.ModelMultipleChoiceFilter(
        field_name="clients__id",
        queryset=Tenant.objects.all(),
        to_field_name="id",
    )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        qs_filter = (
            Q(name__icontains=value) |
            Q(clients__name__icontains=value) 
        )
        return queryset.filter(qs_filter)



    class Meta:
        model = models.Service

        fields = [
            "name",
            "clients",
        ]

class ApplicationFilter(django_filters.FilterSet):
    
    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    devices = django_filters.ModelMultipleChoiceFilter(
        field_name="devices__name",
        queryset=Device.objects.all(),
        to_field_name="name",
    )

    virtual_machines = django_filters.ModelMultipleChoiceFilter(
        field_name="vm__name",
        queryset=VirtualMachine.objects.all(),
        to_field_name="name",
    )

    def search(self, queryset, name, value):

        if not value.strip():
            return queryset
            
        qs_filter = (
            Q(name__icontains=value) |
            Q(devices__name__icontains=value) |
            Q(vm__name__icontains=value)
        )

        return queryset.filter(qs_filter).distinct()



    class Meta:
        model = models.Application

        fields = [
            "name",
            "protocol",
            "version",
        ]


from . import models 
from netbox.views import generic

class ServiceDeleteView(generic.ObjectDeleteView):
    queryset = models.Service.objects.all()

class ApplicationDeleteView(generic.ObjectDeleteView):
    queryset = models.Application.objects.all()



from django.db.models import Q
from utilities.choices import ChoiceSet

OBJETO_ASSIGNMENT_MODELS = Q(
    Q(app_label='dcim', model='device') |
    Q(app_label='circuits', model='circuit') |
    Q(app_label='virtualization', model='virtualmachine') |
    Q(app_label='nb_service', model='application')
)

class PenTestChoices(ChoiceSet):

    APROVED = 1
    REPROVED = 2



    CHOICES = (
        (APROVED, "Approved"),
        (REPROVED, "Reproved"),
    )

class ShapeChoices(ChoiceSet):

    ROUND_EDGES = 1
    STADIUM = 2
    SUBROUTINE = 3
    CYLINDRICAL = 4
    CIRCLE = 5
    ASYMMETRIC = 6
    RHOMBUS = 7
    HEXAGON = 8
    PARALLELOGRAM = 9
    TRAPEZOID = 10
    
    CHOICES = (
        (ROUND_EDGES,  'Round Edges'),
        (STADIUM,  'Stadium Shaped'),
        (SUBROUTINE,  "Subroutine Shape"),
        (CYLINDRICAL,  "Cylindrical Shape"),
        (CIRCLE,  "Circle Shape"),
        (ASYMMETRIC, "asymmetric shape"),
        (RHOMBUS,"rhombus"),
        (HEXAGON,"Hexagon"),
        (PARALLELOGRAM ,"Parallelogram"),
        (TRAPEZOID,"Trapezoid"),
    )

class ConnectorChoices(ChoiceSet):
    ARROW = 1
    OPEN = 2
    DOTTED_ARROW = 3
    DOTTED_OPEN = 4

    CHOICES = (
        (ARROW,  'Arrow'),
        (OPEN,  'Open'),
        (DOTTED_ARROW,  "Dotted Arrow"),
        (DOTTED_OPEN,  "Dotted Open"),
    )


from django.http import HttpResponse
from django.urls import path
from django.conf import settings
from packaging import version


NETBOX_CURRENT_VERSION = version.parse(settings.VERSION)
if NETBOX_CURRENT_VERSION >= version.parse("3.2"):
    from nb_service.views_3_x import  ServiceChangeLogView as ObjectChangeLogView
else:
    from extras.views import ObjectChangeLogView

from . import views
from . import models

urlpatterns = [
    path("service/", views.ServiceListView.as_view(), name="service_list"),
    path("service/<int:pk>/", views.ServiceView.as_view(), name="service"),
    path('service/add/', views.ServiceEditView.as_view(), name='service_add'),
    path('service/<int:pk>/delete/', views.ServiceDeleteView.as_view(), name='service_delete'),
    path('service/<int:pk>/edit/', views.ServiceEditView.as_view(), name='service_edit'),
    path('service/<int:pk>/diagram/', views.ServiceDiagramView.as_view(), name='service_diagram'),
    path('service/<int:pk>/IC/', views.ServiceICView.as_view(), name='service_IC'),
    path('service/<int:pk>/relation/', views.ServiceRelationView.as_view(), name='service_relation'),
    

    path('relation/add', views.RelationEditView.as_view(), name='relation_add'),
    path('relation/<int:pk>/edit/', views.RelationEditView.as_view(), name='relation_edit'),
    path('relation/<int:pk>/delete/', views.RelationDeleteView.as_view(), name='relation_delete'),

    path('ic/add', views.ICCreateView.as_view(), name='IC_add'),
    path('ic/<int:pk>/delete/', views.ICDeleteView.as_view(), name='IC_delete'),

    path('pentest/add', views.PenTestEditView.as_view(), name='pentest_add'),
    path('pentest/<int:pk>/edit/', views.PenTestEditView.as_view(), name='pentest_edit'),
    path('pentest/<int:pk>/delete/', views.PenTestDeleteView.as_view(), name='pentest_delete'),

    path('application/', views.ApplicationListView.as_view(), name='application_list'),
    path("application/<int:pk>/", views.ApplicationView.as_view(), name="application"),
    path("application/<int:pk>/devices", views.ApplicationDevicesView.as_view(), name="application_devices"),
    path("application/<int:pk>/vms", views.ApplicationVMsView.as_view(), name="application_vms"),
    path('application/add', views.ApplicationEditView.as_view(), name='application_add'),
    path('application/<int:pk>/edit/', views.ApplicationEditView.as_view(), name='application_edit'),
    path('application/<int:pk>/delete/', views.ApplicationDeleteView.as_view(), name='application_delete'),
]

urlpatterns_3_2 = []
urlpatterns_2_x = []
NETBOX_CURRENT_VERSION = version.parse(settings.VERSION)
if NETBOX_CURRENT_VERSION >= version.parse("3.2"):
    from nb_service.views_3_x import  ServiceChangeLogView, ApplicationChangeLogView
    urlpatterns_3_2 = [
        path("service/<int:pk>/changelog", ServiceChangeLogView.as_view(), name="service_changelog", kwargs={'model': models.Service}),    
        path("application/<int:pk>/changelog", ApplicationChangeLogView.as_view(), name="application_changelog", kwargs={'model': models.Application}),  
    ]
else:
    from extras.views import ObjectChangeLogView
    urlpatterns_2_x = [
        path("service/<int:pk>/changelog", ObjectChangeLogView.as_view(), name="service_changelog", kwargs={'model': models.Service}),    
        path("application/<int:pk>/changelog", ObjectChangeLogView.as_view(), name="application_changelog", kwargs={'model': models.Application}),  
    ]

urlpatterns.extend(urlpatterns_3_2)
urlpatterns.extend(urlpatterns_2_x)


import django_filters

from utilities.forms import widgets
from django.forms import BoundField
from django import forms
from django.urls import reverse
from django.conf import settings

class DynamicModelChoiceMixin:
    """
    :param display_field: The name of the attribute of an API response object to display in the selection list
    :param query_params: A dictionary of additional key/value pairs to attach to the API request
    :param initial_params: A dictionary of child field references to use for selecting a parent field's initial value
    :param null_option: The string used to represent a null selection (if any)
    :param disabled_indicator: The name of the field which, if populated, will disable selection of the
        choice (optional)
    :param brief_mode: Use the "brief" format (?brief=true) when making API requests (default)
    """
    filter = django_filters.ModelChoiceFilter
    widget = widgets.APISelect

    def __init__(self, display_field='name', query_params=None, initial_params=None, null_option=None,
                 disabled_indicator=None, brief_mode=True, *args, **kwargs):
        self.display_field = display_field
        self.query_params = query_params or {}
        self.initial_params = initial_params or {}
        self.null_option = null_option
        self.disabled_indicator = disabled_indicator
        self.brief_mode = brief_mode

        # to_field_name is set by ModelChoiceField.__init__(), but we need to set it early for reference
        # by widget_attrs()
        self.to_field_name = kwargs.get('to_field_name')

        super().__init__(*args, **kwargs)

    def widget_attrs(self, widget):
        attrs = {
            'display-field': self.display_field,
        }

        # Set value-field attribute if the field specifies to_field_name
        if self.to_field_name:
            attrs['value-field'] = self.to_field_name

        # Set the string used to represent a null option
        if self.null_option is not None:
            attrs['data-null-option'] = self.null_option

        # Set the disabled indicator, if any
        if self.disabled_indicator is not None:
            attrs['disabled-indicator'] = self.disabled_indicator

        # Toggle brief mode
        if not self.brief_mode:
            attrs['data-full'] = 'true'

        # Attach any static query parameters
        for key, value in self.query_params.items():
            widget.add_query_param(key, value)

        return attrs

    def get_bound_field(self, form, field_name):
        bound_field = BoundField(form, self, field_name)

        # Set initial value based on prescribed child fields (if not already set)
        if not self.initial and self.initial_params:
            filter_kwargs = {}
            for kwarg, child_field in self.initial_params.items():
                value = form.initial.get(child_field.lstrip('$'))
                if value:
                    filter_kwargs[kwarg] = value
            if filter_kwargs:
                self.initial = self.queryset.filter(**filter_kwargs).first()

        # Modify the QuerySet of the field before we return it. Limit choices to any data already bound: Options
        # will be populated on-demand via the APISelect widget.
        data = bound_field.value()
        if data:
            field_name = getattr(self, 'to_field_name') or 'pk'
            filter = self.filter(field_name=field_name)
            try:
                self.queryset = filter.filter(self.queryset, data)
            except TypeError:
                # Catch any error caused by invalid initial data passed from the user
                self.queryset = self.queryset.none()
        else:
            self.queryset = self.queryset.none()

        # Set the data URL on the APISelect widget (if not already set)
        widget = bound_field.field.widget
        if not widget.attrs.get('data-url'):
            app_label = self.queryset.model._meta.app_label
            model_name = self.queryset.model._meta.model_name
            data_url = reverse('plugins-api:{}-api:{}-list'.format(app_label, model_name))
            widget.attrs['data-url'] = data_url

        return bound_field


class DynamicModelChoiceField(DynamicModelChoiceMixin, forms.ModelChoiceField):
    """
    Override get_bound_field() to avoid pre-populating field choices with a SQL query. The field will be
    rendered only with choices set via bound data. Choices are populated on-demand via the APISelect widget.
    """

    def clean(self, value):
        """
        When null option is enabled and "None" is sent as part of a form to be submitted, it is sent as the
        string 'null'.  This will check for that condition and gracefully handle the conversion to a NoneType.
        """
        if self.null_option is not None and value == settings.FILTERS_NULL_CHOICE_VALUE:
            return None
        return super().clean(value)


#

# Django imports
from django.conf import settings
from packaging import version
from django.db import models
from django.urls import reverse
from django.contrib.postgres.fields import ArrayField
from django.core.validators import MaxValueValidator, MinValueValidator
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.fields import GenericRelation
from django.core.exceptions import ValidationError


# Netbox imports
from utilities.querysets import RestrictedQuerySet

from dcim.models import Device
from virtualization.models import VirtualMachine
from tenancy.models import Tenant
from ipam.models import Service
from ipam.choices import ServiceProtocolChoices

# version safe netbox imports
NETBOX_CURRENT_VERSION = version.parse(settings.VERSION)
if NETBOX_CURRENT_VERSION >= version.parse("2.11") :
    from netbox.models import ChangeLoggedModel
else:
    from extras.models import ChangeLoggedModel
# module imports
from . import choices

SERVICE_PORT_MIN = 1
SERVICE_PORT_MAX = 65535
SHAPE_NAMES = [ 'Round Edges',
                'Stadium Shaped',
                "Subroutine Shape",
                "Cylindrical Shape",
                "Circle Shape",
                "asymmetric shape",
                "rhombus",
                "Hexagon",
                "Parallelogram",
                "Trapezoid",]

class Service(ChangeLoggedModel):
    name = models.CharField("Name", max_length=100)
    clients = models.ManyToManyField(Tenant,
                related_name="itsm_services",
                verbose_name="Clients")
    comments = models.TextField(
        blank=True
    )
    backup_profile = models.CharField("Backup Profile", max_length=100)

    objects = RestrictedQuerySet.as_manager()

    @property
    def diagram(self):
        graph = "graph TB\n"
        open_shape = [
            '(',
            '([',
            '[[',
            '[(',
            '((',
            '>',
            '{',
            '{{',
            '[/',
            '[/',
            ]

        close_shape = [
            ')',
            '])',
            ']]',
            ')]',
            '))',
            ']',
            '}',
            '}}',
            '/]',
            '\]',
            ]
        arrow_shape = ['-->', '---', '-.->', '-.-']
        nodes = {}
        for rel in self.relationships.all():
            src_node = rel.source.name.replace(" ", "_")
            dest_node = rel.destination.name.replace(" ", "_")
            if src_node not in nodes:
                nodes[src_node] = rel.source.get_absolute_url()
            if dest_node not in nodes:
                nodes[dest_node] = rel.destination.get_absolute_url()
            
            graph += f"    {src_node}{open_shape[rel.source_shape -1]}" + \
                     f"{rel.source.name}{close_shape[rel.source_shape -1]}" + \
                     f" {arrow_shape[rel.connector_shape -1]} "
            if rel.link_text != "":
                graph += f"| {rel.link_text} |"
            graph += f"{dest_node}{open_shape[rel.destination_shape -1]}" + \
                     f"{rel.destination.name}{close_shape[rel.destination_shape -1]}\n"
        for node in nodes:
            graph += f"click {node} \"{nodes[node]}\"\n"
        return graph

    def __str__(self) -> str:
        return f"{self.name}"

    def get_absolute_url(self):
        return reverse("plugins:nb_service:service", kwargs={"pk": self.pk})

class Application(ChangeLoggedModel):
    name = models.CharField("Name", max_length=100)
    protocol = models.CharField("Protocol",
        max_length=50,
        choices=ServiceProtocolChoices
    )
    ports = ArrayField(
        base_field=models.PositiveIntegerField(
            validators=[
                MinValueValidator(SERVICE_PORT_MIN),
                MaxValueValidator(SERVICE_PORT_MAX)
            ]
        ),
        verbose_name='Port numbers'
    )
    version = models.CharField("Version", max_length=100)
    devices = models.ManyToManyField(Device,
                related_name="uses_apps",
                verbose_name="Devices",
                blank=True,)
    vm = models.ManyToManyField(VirtualMachine,
                related_name="uses_apps",
                verbose_name="Virtual Machines",
                blank=True,)
                
    objects = RestrictedQuerySet.as_manager()

    def get_absolute_url(self):
        return reverse("plugins:nb_service:application", kwargs={"pk": self.pk})

    def __str__(self) -> str:
        return f"{self.name} - {self.version}"

class IC(ChangeLoggedModel):
    service = models.ForeignKey(
        to=Service,
        on_delete=models.CASCADE,
        related_name='config_itens'
    )

    assigned_object_type = models.ForeignKey(
        to=ContentType,
        limit_choices_to=choices.OBJETO_ASSIGNMENT_MODELS,
        on_delete=models.CASCADE,
        related_name='+',
        blank=True,
        null=True
    )
    assigned_object_id = models.PositiveIntegerField(
        blank=True,
        null=True
    )
    assigned_object = GenericForeignKey(
        ct_field='assigned_object_type',
        fk_field='assigned_object_id',
    )

    objects = RestrictedQuerySet.as_manager()
    
    @property
    def obj_type(self):
        return self.assigned_object_type.name

    @property
    def name(self):
        if self.assigned_object is not None:
            return self.assigned_object.name
        else:
            return ""

    def __str__(self):
        return str(self.assigned_object)

    def get_absolute_url(self):
        url = reverse("plugins:nb_service:service", kwargs={"pk": self.service.pk})
        try:
            url = self.assigned_object.get_absolute_url()
        except Exception as e:
            pass
        return url
    
    def validate_unique(self, exclude=None):
        ICs = self.service.config_itens.filter(assigned_object_type=self.assigned_object_type)
        candidate_obj = self.assigned_object
        for obj in ICs:
            if candidate_obj == obj.assigned_object:
                raise ValidationError((f"{self.assigned_object} already in {self.service}"))

        super().validate_unique(exclude=exclude)
    
    class Meta:
        unique_together = [
             ['service', 'assigned_object_type', 'assigned_object_id'],
        ]

    
class Relation(ChangeLoggedModel):
    service = models.ForeignKey(
        verbose_name="Service",
        to=Service,
        on_delete=models.CASCADE,
        related_name='relationships'
    )
    source = models.ForeignKey(
        verbose_name="Source",
        to=IC,
        on_delete=models.CASCADE,
        related_name='+'
    )
    source_shape = models.IntegerField("Source Shape",
        choices=choices.ShapeChoices
    )
    destination = models.ForeignKey(
        verbose_name="Destination",
        to=IC,
        on_delete=models.CASCADE,
        related_name='+'
    )
    destination_shape = models.IntegerField("Destination Shape",
        choices=choices.ShapeChoices
    )
    connector_shape = models.IntegerField("Connector Shape",
        choices=choices.ConnectorChoices
    )
    
    link_text = models.CharField("Link Text", max_length=100)

    objects = RestrictedQuerySet.as_manager()
    
    @property
    def source_shape_name(self):        
        return SHAPE_NAMES[self.source_shape-1]    

    @property
    def destiny_shape_name(self):        
        return SHAPE_NAMES[self.destination_shape-1]    

    @property
    def connector(self):
        arrow_shape = ['-->', '---', '-.->', '-.-']
        return f"{arrow_shape[self.connector_shape -1]}"

    def __str__(self) -> str:
        arrow_shape = ['-->', '---', '-.->', '-.-']
        src_node = self.source.name.replace(" ", "_")
        dest_node = self.destination.name.replace(" ", "_")

        return f"{src_node} {arrow_shape[self.connector_shape -1]} {dest_node}"


class PenTest(ChangeLoggedModel):
    service = models.ForeignKey(
        to=Service,
        on_delete=models.CASCADE,
        related_name='pentest_reports'
    )
    comments = models.TextField(
        blank=True
    )
    status = models.IntegerField(
        "State",
        choices=choices.PenTestChoices,
        blank=True,
        help_text='Approved or Reproved'
    )
    date = models.DateField("Execution Date", blank=True)
    ticket = models.CharField("Ticket", max_length=100)
    report_link = models.CharField("Report link", max_length=100)

    objects = RestrictedQuerySet.as_manager()
